@@ -17,7 +17,8 @@ class Viewer(App):
     A thin wrapper on :class:`compas_view2.app.App`.
     """
     def __init__(self, *args, **kwargs):
-        super(Viewer, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
+        self.artists = []
 
     def add(self, data, **kwargs):
         """
@@ -45,5 +46,6 @@ class Viewer(App):
             return super(Viewer, self).add(data, **kwargs)
 
         artist = Artist(data, viewer=self, context="Viewer", **kwargs)
-
-        return artist.draw()
+        self.artists.append(artist)
+        artist.draw()
+        artist.add()
